
#property copyright "Copyright 2011, TheXpert"

class CallInfo
{
public:   
   CallInfo()
      : m_File("")
      , m_Line(0)
      , m_Function("")
      , m_Comment("")
   {
   }
   
   CallInfo(string file, uint line, string function, string comment)
      : m_File(file)
      , m_Line(line)
      , m_Function(function)
      , m_Comment(comment)
   {
   }
   
   CallInfo(const CallInfo& call)
      : m_File(call.m_File)
      , m_Line(call.m_Line)
      , m_Function(call.m_Function)
      , m_Comment(call.m_Comment)
   {
   }
   
   bool CopyFrom(CallInfo* call)
   {
      if (CheckPointer(call) == POINTER_INVALID) return false;
      
      m_File = call.m_File;
      m_Line = call.m_Line;
      m_Function = call.m_Function;
      m_Comment = call.m_Comment;
      
      return true;
   }
   
public:   
   string m_File;
   uint m_Line;
   string m_Function;
   string m_Comment;
};

class Call
   : public CallInfo
{
public:
   Call(string file, uint line, string function, string comment)
      : CallInfo(file, line, function, comment)
   {
   }
   
   ~Call()
   {
      MacroImplGeHelper().Detach();
   }
};

class Tracer
{
public:
   bool Last(CallInfo& call)
   {
      int last = ArraySize(m_CallStack) - 1;
      m_Pos = last;
      if (last < 0) return false;
      
      return call.CopyFrom(m_CallStack[last]);
   }

   bool Previous(CallInfo& call)
   {
      if (m_Pos < 0)
      {
         int last = ArraySize(m_CallStack) - 1;
         m_Pos = last;
         if (last < 0) return false;
      }
      
      m_Pos--;
      if (m_Pos < 0) return false;
      
      return call.CopyFrom(m_CallStack[m_Pos]);
   }
   
   void Stack(CallInfo*& stack[])
   {
      ArrayCopy(stack, m_CallStack);
   }
   
   void SaveToFile(string filePath)
   {
      int hFile = FileOpen(filePath, FILE_CSV | FILE_WRITE);
      if (hFile == -1) return;

      int size = ArraySize(m_CallStack);
      string s;
      
      for (int i = size - 1; i >= 0; i--)      
      {
         s = m_CallStack[i].m_File + "  at  line " + string(m_CallStack[i].m_Line) + "  function: <<" + m_CallStack[i].m_Function + ">>";
         if (m_CallStack[i].m_Comment != "")
         {
            s = s + "  with  comment \"" + m_CallStack[i].m_Comment + "\"";
         }
         FileWrite(hFile, s);
      }
      
      FileClose(hFile);
   }

protected:
   // class should not be used separately. Use macros
   Tracer()
      : m_Pos(-1)
   {}
   
   CallInfo* m_CallStack[];
   int m_Pos;
};

class TraceHelper
   : public Tracer
{
public:
   void Attach(const CallInfo& call)
   {
      uint size = ArraySize(m_CallStack);
      ArrayResize(m_CallStack, size + 1);
      m_CallStack[size] = new CallInfo(call);
   }
   
   void Detach()
   {
      uint size = ArraySize(m_CallStack) - 1;
      delete m_CallStack[size];
      ArrayResize(m_CallStack, size);
   }

   TraceHelper()
      : Tracer()
   {}
};

TraceHelper* MacroImplGeHelper()
{
   static TraceHelper TheHelper;
   return GetPointer(TheHelper);
}

Tracer* GetTracer()
{
   return MacroImplGeHelper();
}