//+------------------------------------------------------------------+
//|                                                FileUnlimited.mq5 |
//+------------------------------------------------------------------+
#property library
#property copyright "TheXpert"
#property link "theforexpert@gmail.com"
#property version   "1.00"

#include <TheXpert/FileUnlimited/Interfaces.mqh>
#include <TheXpert/StringUtils.mqh>

#include <TheXpert/DebugTools/ASSERT.mqh>
#include <TheXpert/DebugTools/Trace/Trace.mqh>

#import "kernel32.dll"
   bool   CloseHandle(int hObject);
   int    CreateFileW(     string lpFileName,         uint dwDesiredAccess, 
                           uint dwShareMode,          uint lpSecurityAttributes,
                           uint dwCreationDisposition,uint dwFlagsAndAttributes, 
                           int hTemplateFile
                     );
   uint   GetFileAttributesW(string lpFileName);
   uint   GetFileSize(int hFile, uint& lpFileSizeHigh[]);
   bool   GetFileTime(int hFile, uint& lpCreationTime[], uint& lpLastAccessTime[], uint& lpLastWriteTime[]);
   bool   ReadFile(int hFile, ushort& lpBuffer[], uint nNumberOfBytesToRead, uint& lpNumberOfBytesRead[], int lpOverlapped);
   uint   SetFilePointer(int hFile, int lDistanceToMove, int& lpDistanceToMoveHigh[], uint dwMoveMethod);
   bool   FileTimeToSystemTime(const uint& lpFileTime[], MqlDateTime& lpSystemTime[]);

   int    SetEndOfFile (int hFile);
   bool   WriteFile (int  hFile, ushort& lpBuffer[], int  nNumberOfBytesToWrite, int& lpNumberOfBytesWritten[], int  lpOverlapped);
#import
//===========================================================

class FileImpl
   : public FileWide
{
public:
   FileImpl(string path, CreationDisposition options, DesiredAccess access, ushort writeSeparator, string readSeparators);
   
   bool IsValid() const 
   {
      return m_hFile != INVALID_HANDLE;
   }
   
   ~FileImpl()
   {
      if (m_hFile != INVALID_HANDLE)
      {
         CloseHandle(m_hFile);
      }
   }
   
   virtual ulong Size() const 
   {
      return m_Size;
   }
   
   virtual string Path() const
   {
      return m_Path;
   }

   virtual datetime CreationTime() const 
   {
      return m_CreationTime;
   }
   
   virtual datetime LastAccessTime() const 
   {
      return m_LastAccessTime;
   }
   
   virtual datetime LastModifyTime() const 
   {
      return m_LastModifyTime;
   }
   
   virtual bool Read(bool& value, string separators);
   virtual bool Read(long& value, string separators);
   virtual bool Read(int& value, string separators);
   virtual bool Read(double& value, string separators);
   virtual bool Read(string& value, string separators);
   virtual bool Read(datetime& value, string separators);
   virtual bool Read(color& value, string separators);

   virtual bool Read(bool& value);
   virtual bool Read(long& value);
   virtual bool Read(int& value);
   virtual bool Read(double& value);
   virtual bool Read(string& value);
   virtual bool Read(datetime& value);
   virtual bool Read(color& value);

   virtual bool SetPos(long pos, MovePointerMethod from = MPM_BEGIN);
   virtual string ToString();

   virtual long Find(string what);
   virtual bool IsEnding() const;
   
   virtual bool Write(bool value);
   virtual bool Write(long value);
   virtual bool Write(int value);
   virtual bool Write(double value);
   virtual bool Write(string value);
   virtual bool Write(datetime value);
   virtual bool Write(color value);

   virtual bool Write(bool value, ushort separator);
   virtual bool Write(long value, ushort separator);
   virtual bool Write(int value, ushort separator);
   virtual bool Write(double value, ushort separator);
   virtual bool Write(string value, ushort separator);
   virtual bool Write(datetime value, ushort separator);
   virtual bool Write(color value, ushort separator);

   virtual bool Insert(string what, long pos, MovePointerMethod from);
   virtual bool Insert(string what);

   virtual bool Delete(ulong count, long pos, MovePointerMethod from);
   virtual bool Delete(ulong count);

private:
   void Init();
   bool ReadBlock(uint size = 1000);
   
   bool ReadString(string& value, string separators);
   bool WriteString(const string& value);
   
   void FixEndOfLine();
   
   void ReadTimeOptions();

   bool Resize(ulong newSize);
private:
   int m_hFile;
   ulong m_Size;
   datetime m_CreationTime;
   datetime m_LastAccessTime;
   datetime m_LastModifyTime;
   
   string m_Buffer;
   ulong m_Pos;
   uint m_Attributes;
   string m_Path;
   
   bool m_Modified;
   ushort m_WriteSeparator;
   string m_ReadSeparators;
};

FileImpl::FileImpl(string path, CreationDisposition options, DesiredAccess access, ushort writeSeparator, string readSeparators)
   : FileWide(path, options, access)
   , m_hFile(INVALID_HANDLE)
   , m_Size(0)
   , m_CreationTime(0)
   , m_LastAccessTime(0)
   , m_LastModifyTime(0)
   , m_Buffer("")
   , m_Pos(0)
   , m_Attributes(INVALID_FILE_ATTRIBUTES)
   , m_Path(path)
   , m_Modified(false)
   , m_WriteSeparator(writeSeparator)
   , m_ReadSeparators(readSeparators)
{__TRACE
   m_hFile = CreateFileW( path, FU_GENERIC_WRITE, FU_FILE_SHARE_WRITE, 0, options, 0, 0);
   
   ASSERT(m_hFile != INVALID_HANDLE);
   if (m_hFile != INVALID_HANDLE)
   {
      Init();
   }
}

void FileImpl::ReadTimeOptions()
{__TRACE
   // requesting time properties
   uint creation[4], lastAccess[4], lastModify[4];
   if (GetFileTime(m_hFile, creation, lastAccess, lastModify))
   {
      // request succeeded, need to tranform time
      MqlDateTime time[1];
      FileTimeToSystemTime(creation, time);
      m_CreationTime = StructToTime(time[0]);

      FileTimeToSystemTime(lastAccess, time);
      m_LastAccessTime = StructToTime(time[0]);

      FileTimeToSystemTime(lastModify, time);
      m_LastModifyTime = StructToTime(time[0]);
   }
   else ASSERT(false);
}

void FileImpl::Init()
{__TRACE
   // requesting size
   uint hi[1] = {0};
   uint low = GetFileSize(m_hFile, hi);
   
   m_Size = ulong(hi);
   m_Size = m_Size >> 32;
   m_Size += low;
   
   ASSERT(m_Size % 2 == 0);
   
   ReadTimeOptions();
   
   m_Attributes = GetFileAttributesW(m_Path);
}

bool FileImpl::Read(bool& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(long& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(int& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(double& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(string& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(datetime& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(color& value)
{
   return Read(value, m_ReadSeparators);
}

bool FileImpl::Read(bool& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToBool(tmp);
      return true;
   }
   return false;
}

bool FileImpl::Read(long& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToInteger(tmp);
      return true;
   }
   return false;
}

bool FileImpl::Read(int& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = int(StringToInteger(tmp));
      return true;
   }
   return false;
}

bool FileImpl::Read(double& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToDouble(tmp);
      return true;
   }
   return false;
}

bool FileImpl::ReadString(string& value, string separators)
{__TRACE
   if (m_Buffer == "")
   {
      if (!ReadBlock()) return false;
   }
   
   int pos = StringFindAnyOfChar(m_Buffer, separators);
   while (pos == 0)
   {
      m_Buffer = StringSubstr(m_Buffer, 1);
      m_Pos += 2;
      pos = StringFindAnyOfChar(m_Buffer, separators);
   }
   
   string strValue = "";
   
   bool needFixEOL = false;
   
   if (pos == -1)
   {
      strValue = m_Buffer;
      m_Pos += 2*StringLen(m_Buffer);
      m_Buffer = "";
      
      bool succeeded = ReadBlock();
      
      while (succeeded)
      {
         if (strValue == "")
         {
            pos = StringFindAnyOfChar(m_Buffer, separators);
            while (pos = 0)
            {
               m_Buffer = StringSubstr(m_Buffer, 1);
               m_Pos += 2;
            }
            
            if (pos == -1)
            {
               strValue = m_Buffer;
               m_Pos += 2*StringLen(m_Buffer);
               m_Buffer = "";
            }
            else
            {
               strValue = StringSubstr(m_Buffer, 0, pos);

               ushort chr = StringGetCharacter(m_Buffer, pos);
               if (chr == '\r') needFixEOL = true;

               m_Pos += 2*(StringLen(strValue) + 1);
               m_Buffer = StringSubstr(m_Buffer, pos + 1);
               break;
            }
         }
         else
         {
            pos = StringFindAnyOfChar(m_Buffer, separators);
            if (pos == 0)
            {
               ushort chr = StringGetCharacter(m_Buffer, pos);
               if (chr == '\r') needFixEOL = true;

               m_Buffer = StringSubstr(m_Buffer, 1);
               m_Pos += 2;
               break;
            }
            else if (pos == -1)
            {
               strValue = m_Buffer;
               m_Pos += 2*StringLen(m_Buffer);
               m_Buffer = "";
            }
            else
            {
               strValue = strValue + StringSubstr(m_Buffer, 0, pos);

               ushort chr = StringGetCharacter(m_Buffer, pos);
               if (chr == '\r') needFixEOL = true;

               m_Pos += 2*(StringLen(strValue) + 1);
               m_Buffer = StringSubstr(m_Buffer, pos + 1);
               break;
            }
         }
         succeeded = ReadBlock();
      }

      if (strValue == "") return false;
   }
   else
   {
      strValue = StringSubstr(m_Buffer, 0, pos);

      ushort chr = StringGetCharacter(m_Buffer, pos);
      if (chr == '\r') needFixEOL = true;

      m_Pos += 2*(StringLen(strValue) + 1);
      m_Buffer = StringSubstr(m_Buffer, pos + 1);
   }
   
   if (needFixEOL) FixEndOfLine();
   
   value = strValue;
   
   return true;
}

bool FileImpl::Read(string& value, string separators)
{__TRACE
   if (m_Buffer == "")
   {
      if (!ReadBlock()) return false;
   }
   
   bool needFixEOL = false;
   
   int pos = StringFindAnyOfChar(m_Buffer, separators);
   if (pos == 0)
   {
      ushort chr = StringGetCharacter(m_Buffer, pos);
      if (chr == '\r') needFixEOL = true;

      m_Buffer = StringSubstr(m_Buffer, 1);
      m_Pos += 2;
      value = "";
      
      if (needFixEOL) FixEndOfLine();
      
      return true;
   }
   
   string strValue = "";
   
   if (pos == -1)
   {
      strValue = m_Buffer;
      m_Pos += 2*StringLen(m_Buffer);
      m_Buffer = "";
      
      bool succeeded = ReadBlock();
      
      while (succeeded)
      {
         pos = StringFindAnyOfChar(m_Buffer, separators);
         if (pos == 0)
         {
            ushort chr = StringGetCharacter(m_Buffer, pos);
            if (chr == '\r') needFixEOL = true;

            m_Buffer = StringSubstr(m_Buffer, 1);
            m_Pos += 2;
            break;
         }
         
         if (pos == -1)
         {
            strValue = m_Buffer;
            m_Pos += 2*StringLen(m_Buffer);
            m_Buffer = "";
         }
         else
         {
            ushort chr = StringGetCharacter(m_Buffer, pos);
            if (chr == '\r') needFixEOL = true;

            strValue = StringSubstr(m_Buffer, 0, pos);
            m_Pos += 2*(StringLen(strValue) + 1);
            m_Buffer = StringSubstr(m_Buffer, pos + 1);
            break;
         }

         succeeded = ReadBlock();
      }
   }
   else
   {
      ushort chr = StringGetCharacter(m_Buffer, pos);
      if (chr == '\r') needFixEOL = true;

      strValue = StringSubstr(m_Buffer, 0, pos);
      m_Pos += 2*(StringLen(strValue) + 1);
      m_Buffer = StringSubstr(m_Buffer, pos + 1);
   }
   
   if (needFixEOL) FixEndOfLine();
   
   value = strValue;
   
   return true;
}

bool FileImpl::Read(datetime& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = datetime(StringToInteger(tmp));
      return true;
   }
   return false;
}

bool FileImpl::Read(color& value, string separators)
{__TRACE
   string tmp;
   if (ReadString(tmp, separators))
   {
      value = StringToColor(tmp);
      return true;
   }
   return false;
}

bool FileImpl::SetPos(long pos, MovePointerMethod from)
{__TRACE
   ASSERT(pos % 2 == 0); // Wide chars have size == 2 bytes. Possible logical error
   ulong newPos = 0;
   
   switch (from)
   {
      case MPM_BEGIN:
         ASSERT(pos >= 0 && pos <= long(m_Size));
         if (pos < 0 || pos > long(m_Size)) return false;
         newPos = pos;
         break;
         
      case MPM_CURRENT:
         ASSERT(pos + m_Pos >= 0 && ulong(pos + m_Pos) <= m_Size);
         if (long(pos + m_Pos) < 0 || ulong(pos + m_Pos) > m_Size) return false;
         newPos = pos + m_Pos;
         break;
         
      case MPM_END:
         ASSERT(m_Size + pos >= 0 && pos <= 0);
         if (long(m_Size + pos) < 0 || pos > 0) return false;
         newPos = m_Size + pos;
         break;
         
      default:
         ASSERT(false);
         return false;
   }
   
   m_Pos = newPos;
   m_Buffer = "";
   
   return true;
}

bool FileImpl::ReadBlock(uint size)
{__TRACE
   ASSERT(m_Pos % 2 == 0); // Wide chars have size == 2 bytes. Possible logical error

   if (m_Pos >= m_Size)
   {
      m_Buffer = "";
      return false;
   }
   
   int lo = int (m_Pos & 0xFFFFFFFF);
   int hi[1];
   hi[0] = int ((m_Pos >> 32) & 0xFFFFFFFF);
   
   ulong newPos = SetFilePointer(m_hFile, lo, hi, MPM_BEGIN);
   if (newPos == INVALID_SET_FILE_POINTER)
   {
      ASSERT(false);
      return false;
   }
   newPos = newPos + (ulong(hi[0] << 32));

   if (newPos != m_Pos)
   {
      ASSERT(false);
      return false;
   }
   
   if (m_Pos >= m_Size - 1)
   {
      m_Buffer = "";
      return false;
   }
   
   int readBytes[] = {0};
   uint toRead = size;
   if (m_Pos + toRead > m_Size)
   {
      toRead = uint(m_Size - m_Pos);
   }

   ushort buffer[];
   ArrayResize(buffer, (toRead + 1)/2);
   
   if (ReadFile(m_hFile, buffer, toRead, readBytes, NULL))
   {
      uint newBufSize = (readBytes[0] + 1)/2;
      
      StringInit(m_Buffer, newBufSize, ' ');
      for (uint i = 0; i < newBufSize; i++)
      {
         StringSetCharacter(m_Buffer, i, buffer[i]);
      }
   }
   else ASSERT(false);
   
   return true;
}

string FileImpl::ToString()
{__TRACE
   m_Buffer = "";
   
   int hi[] = {0};

   uint newPos = SetFilePointer(m_hFile, 0, hi, MPM_BEGIN);

   if (newPos == INVALID_SET_FILE_POINTER)
   {
      ASSERT(false);
      return "";
   }
   
   int readBytes[] = {0};

   ushort buffer[];
   ArrayResize(buffer, int((m_Size + 1)/2));
   
   if (ReadFile(m_hFile, buffer, int(m_Size), readBytes, NULL))
   {
      uint newBufSize = (readBytes[0] + 1)/2;
      
      string result;
      StringInit(result, newBufSize, ' ');
      for (uint i = 0; i < newBufSize; i++)
      {
         StringSetCharacter(result, i, buffer[i]);
      }
      
      return result;
   }
   else ASSERT(false);
   
   return "";
}

void FileImpl::FixEndOfLine(void)
{__TRACE
   if (StringLen(m_Buffer) == 0)
   {
      if (!ReadBlock()) return;
   }
   
   ushort chr = StringGetCharacter(m_Buffer, 0);
   if (chr == '\n')
   {
      m_Pos += 2;
      m_Buffer = StringSubstr(m_Buffer, 1);
   }
}

long FileImpl::Find(string what)
{__TRACE
   long result = -1;
   
   ulong oldPos = m_Pos;
   
   m_Buffer = "";
   m_Pos = 0;
   
   uint length = StringLen(what);
   uint size = 2*length + 1000;
   
   bool succeeded = ReadBlock(size);
   
   while (succeeded)
   {
      result = StringFind(m_Buffer, what);
      
      if (result != -1) break;
      
      m_Pos += length + 1000;
      m_Buffer = "";
      succeeded = ReadBlock(size);
   }
   
   m_Pos = oldPos;
   m_Buffer = "";
   
   return result;
}

bool FileImpl::IsEnding() const
{
   return m_Pos >= m_Size;
}

bool FileImpl::Resize(ulong newSize)
{__TRACE
   int lo = int (newSize & 0xFFFFFFFF);
   int hi[1];
   hi[0] = int ((newSize >> 32) & 0xFFFFFFFF);
   
   ulong newPos = SetFilePointer(m_hFile, lo, hi, MPM_BEGIN);
   if (newPos == INVALID_SET_FILE_POINTER)
   {
      ASSERT(false);
      return false;
   }
   newPos = newPos + (ulong(hi[0] << 32));

   ASSERT (newPos == newSize);
   if (newPos != newSize) return false;
   
   if (SetEndOfFile(m_hFile) == 0)
   {
      ASSERT(false);
      return(false);
   }
   
   m_Size = newSize;
   
   return (true);
}


bool FileImpl::WriteString(const string& value)
{__TRACE
   m_Buffer = "";
   
   uint len = StringLen(value);
   ulong newSize = m_Pos + 2*len;
   
   if (newSize > m_Size)
   {
      if (!Resize(newSize))
      {
         ASSERT(false);
         return false;
      }
   }
   
   int lo = int (m_Pos & 0xFFFFFFFF);
   int hi[1];
   hi[0] = int ((m_Pos >> 32) & 0xFFFFFFFF);
   
   ulong newPos = SetFilePointer(m_hFile, lo, hi, MPM_BEGIN);
   if (newPos == INVALID_SET_FILE_POINTER)
   {
      ASSERT(false);
      return false;
   }
   newPos = newPos + (ulong(hi[0] << 32));
   
   if (newPos != m_Pos)
   {
      ASSERT(false);
      return false;
   }

   ushort buffer[];
   ArrayResize(buffer, len);
   for (uint i = 0; i < len; i++)
   {
      buffer[i] = StringGetCharacter(value, i);
   }
   
   int written[] = {0};
   
   if (!WriteFile(m_hFile, buffer, 2*len, written, 0))
   {
      ASSERT(false);
      return false;
   }
   else
   {
      m_Pos += written[0];
   }
   
   return true;
}

bool FileImpl::Write(bool value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(long value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(int value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(double value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(string value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(datetime value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(color value)
{
   return Write(value, m_WriteSeparator);
}

bool FileImpl::Write(bool value, ushort separator)
{__TRACE
   string s = string(value);
   StringSetCharacter(s, StringLen(s), separator);
   
   return WriteString(s);
}

bool FileImpl::Write(long value, ushort separator)
{__TRACE
   string s = string(value);
   StringSetCharacter(s, StringLen(s), separator);

   return WriteString(s);
}

bool FileImpl::Write(int value, ushort separator)
{__TRACE
   string s = string(value);
   StringSetCharacter(s, StringLen(s), separator);

   return WriteString(s);
}

bool FileImpl::Write(double value, ushort separator)
{__TRACE
   string s = string(value);
   StringSetCharacter(s, StringLen(s), separator);

   return WriteString(s);
}

bool FileImpl::Write(string value, ushort separator)
{__TRACE
   string s = value;
   StringSetCharacter(s, StringLen(s), separator);

   return WriteString(s);
}

bool FileImpl::Write(datetime value, ushort separator)
{__TRACE
   string s = IntegerToString(value);
   StringSetCharacter(s, StringLen(s), separator);
   
   return WriteString(s);
}

bool FileImpl::Write(color value, ushort separator)
{__TRACE
   string s = ColorToString(value);
   StringSetCharacter(s, StringLen(s), separator);

   return WriteString(s);
}

bool FileImpl::Insert(string what, long pos, MovePointerMethod from)
{__TRACE
   ulong insPos = 0;
   switch (from)
   {
      case MPM_BEGIN:   insPos = pos; break;
      case MPM_CURRENT: insPos = m_Pos + pos; break;
      case MPM_END:     insPos = m_Size - 1 + pos; break;
      default: ASSERT(false);
   }
   
   ASSERT(insPos <= m_Size);
   if (insPos > m_Size) return false;
   
   if (what == "") return true;
   
   ulong insSize = 2*StringLen(what);
   ulong newSize = m_Size + insSize;
   ASSERT(newSize % 2 == 0);
   
   if (!Resize(newSize))
   {
      ASSERT(false);
      return false;
   }
   
   if (insPos != m_Size)
   {
      uint blockSize = 5000;
      
      ulong start = m_Size;
      uint size = 0;
      
      while (start > insPos)
      {
         size = uint(start - insPos);
         if (size > blockSize) size = blockSize;
         
         start -= size;
         
         if (!SetPos(start - insSize))
         {
            ASSERT(false);
            return false;
         }
         if (!ReadBlock(size))
         {
            ASSERT(false);
            return false;
         }
         if (!SetPos(start))
         {
            ASSERT(false);
            return false;
         }
         if (!WriteString(m_Buffer))
         {
            ASSERT(false);
            return false;
         }
      }
   }
   
   if (!SetPos(insPos))
   {
      ASSERT(false);
      return false;
   }
   if (!WriteString(what))
   {
      ASSERT(false);
      return false;
   }
   
   return true;
}

bool FileImpl::Delete(ulong count, long pos, MovePointerMethod from)
{__TRACE
   if (count == 0) return true;
   
   ulong delPos = 0;
   switch (from)
   {
      case MPM_BEGIN:   delPos = pos; break;
      case MPM_CURRENT: delPos = m_Pos + pos; break;
      case MPM_END:     delPos = m_Size - 1 + pos; break;
      default: ASSERT(false);
   }
   
   ASSERT(delPos <= m_Size);
   if (delPos > m_Size) return false;
   
   ulong movePos = delPos + 2*count;
   
   if (movePos < m_Size)
   {
      uint blockSize = 5000;
      
      ulong start = movePos;
      uint size = 0;
      
      while (start < m_Size)
      {
         size = uint(m_Size - start);
         if (size > blockSize) size = blockSize;
         
         if (!SetPos(start))
         {
            ASSERT(false);
            return false;
         }
         if (!ReadBlock(size))
         {
            ASSERT(false);
            return false;
         }
         if (!SetPos(start - 2*count))
         {
            ASSERT(false);
            return false;
         }
         if (!WriteString(m_Buffer))
         {
            ASSERT(false);
            return false;
         }

         start += size;
      }
   }
   
   if (!Resize(m_Size - 2*count))
   {
      ASSERT(false);
      return false;
   }

   return true;
}

bool FileImpl::Insert(string what)
{
   return Insert(what, m_Pos, MPM_BEGIN);
}

bool FileImpl::Delete(ulong count)
{
   return Delete(count, m_Pos, MPM_BEGIN);
}

FileWide* OpenFileWide(string path, CreationDisposition options, DesiredAccess access, ushort writeSeparator = ';', string readSeparators = " ;\r\n\t") export
{__TRACE
   FileImpl* result = new FileImpl(path, options, access, writeSeparator, readSeparators);
   
   if (!result.IsValid())
   {
      ASSERT(false);
      delete result;
      result = NULL;
   }
   
   return result;
}